// Copyright 2016 the Go-FUSE Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Package test holds the tests for Go-FUSE and is not for end-user
// consumption.
package test
